/*
 *  Copyright 2001 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Aapl.
 *
 *  Aapl is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Aapl is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Aapl; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#ifndef _AAPL_GRAPH_CC
#define _AAPL_GRAPH_CC

#include <stdio.h>
#include "graph.h"


/****************************************
 * Graph Copy Constructor
 */
template < class State, class Key, class Transition >
		Graph<State, Key, Transition>::Graph(Graph &graph) :
		LastTransElFrom(0), LastTransElTo(0)
{
	/* Create the states and record their map in the original state. */
	State *origState = graph.Head;
	while ( origState != 0 ) {
		/* Make the new state. */
		State *newState = new State(*origState);

		/* Add the state to the  */
		AddToGraph( newState );

		/* Set the mapsTo item of the old state. */
		origState->stateMap = newState;

		/* Next state. */
		origState = origState->Next;
	}
	
	/* Derefernce all the state maps and resort. */
	State *state = Head;
	while ( state != 0 ) {
		/* Walk the list of out and in transitions setting the toState values
		 * to their mapped values into the set of new states. */
		typename State::TransitionEl *el = state->InList.Table;
		int len = state->InList.TableLength;
		for (int i = 0; i < len; i++, el++ )
			el->Key.toState = el->Key.toState->stateMap;

		el = state->OutList.Table;
		len = state->OutList.TableLength;
		for (int i = 0; i < len; i++, el++ )
			el->Key.toState = el->Key.toState->stateMap;

		/* Re-sort the in and out trans lists. Remember they are bstables
		 * paritally keyed on the toState, which has changed. */
		MergeSort<typename State::TransitionEl,
				typename State::TransitionElCompare>::
				Sort (state->InList.Table, state->InList.TableLength );
		MergeSort<typename State::TransitionEl,
				typename State::TransitionElCompare>::
				Sort(state->OutList.Table, state->OutList.TableLength );

		state = state->Next;
	}
}

template < class State, class Key, class Transition>
		Graph<State, Key, Transition>::~Graph()
{
	/* Go through all the transitions and delete. */
	State *state = Head;
	while ( state != 0 ) {
		typename State::TransitionEl *tel = state->OutList.Table;
		int tlen = state->OutList.TableLength;
		for ( int i = 0; i < tlen; i++, tel++ ) {
			if ( tel->Value != 0 )
				delete tel->Value;
		}
		state = state->Next;
	}
}

template < class State, class Key, class Transition >
	bool Graph<State, Key, Transition>::AttachStates(
			State *from, State *to, Key onChar)
{
	bool retVal;
	if ( from->OutList.Insert(
				typename State::TransitionKey(onChar, to), 0, &LastTransElFrom) ) {
		/* Adding the trans in the out list succeeded. Add it in the in trans. */
		to->InList.Insert(typename State::TransitionKey(onChar, from), 0, &LastTransElTo);
		retVal = true;
	}
	else {
		/* Trans is already there. Do a lookup in the inlist so we can
		 * have a last found. If it is not there then we have a broken graph. */
		LastTransElTo = to->InList.Find( typename State::TransitionKey(onChar, from) );
		ASSERT( LastTransElTo != NULL );
		retVal = false;
	}

	return retVal;
}

template < class State, class Key, class Transition >
	Transition *Graph<State, Key, Transition>::AttachStatesWithTrans(
			State *from, State *to, Key onChar)
{
	if ( AttachStates( from, to, onChar ) ) {
		Transition *newTrans = new Transition;
		SetLastTrans( newTrans );
		return newTrans;
	}
	else
		return GetLastTrans();
}

template < class State, class Key, class Transition >
	Transition *Graph<State, Key, Transition>::DetachStates(
			State *from, State *to, Key onChar )
{
	Transition *retTrans = 0;

	/* Look for the transition key. */
	typename State::TransitionEl *transEl = 
			from->OutList.Find( typename State::TransitionKey(onChar, to) );
	if ( transEl != NULL ) {
		/* Save the Transition. */
		retTrans = transEl->Value;

		/* Remove from OutList. */
		from->OutList.RemoveItem( transEl );
	
		/* Remove from InList. */
		to->InList.Remove( typename State::TransitionKey(onChar, from) );
	}

	return retTrans;
}

template < class State, class Key, class Transition>
		State *Graph<State, Key, Transition>
		::DetachState(State *state)
{
	DetachInTrans(state);
	DetachOutTrans(state);
	DoubleList<State>::DetachElement(state);
	return state;
}

template < class State, class Key, class Transition>
		State *Graph<State, Key, Transition>
		::DetachInTrans(State *state)
{
	/* Remove the outgoings assoc the incommings. */
	typename State::TransitionEl *tel = state->InList.Table;
	int tlen = state->InList.TableLength;
	for (int i = 0; i < tlen; i++, tel++)
	{
		tel->Key.toState->OutList.Remove(
				typename State::TransitionKey(tel->Key.onChar, state));
		if ( tel->Value != 0 )
			delete tel->Value;
	}

	state->InList.Empty();
	return state;
}

template < class State, class Key, class Transition >
		State *Graph<State, Key, Transition>
		::DetachOutTrans(State *state)
{
	/* Remove the incommings assoc the outgoings. */
	typename State::TransitionEl *tel = state->OutList.Table;
	int tlen = state->OutList.TableLength;
	for (int i = 0; i < tlen; i++, tel++)
	{
		tel->Key.toState->InList.Remove(
				typename State::TransitionKey(tel->Key.onChar, state));
		if ( tel->Value != 0 )
			delete tel->Value;
	}

	state->OutList.Empty();
	return state;
}

/****************************************
 * Number the states starting at 0.
 */
template < class State, class Key, class Transition >
		void Graph<State, Key, Transition>::SetStateNumbers()
{
	int curNum = 0;
	State *state = Head;
	while (state != 0) {
		state->Num = curNum;
		curNum++;
		state = state->Next;
	}
}

template < class State, class Key, class Transition >
		void Graph<State, Key, Transition>::
		DumpTransitions( FILE *file, typename State::TransitionList *list )
{
	typename State::TransitionEl *tel = list->Table;
	int tlen = list->TableLength;
	for ( int i = 0; i < tlen; i++, tel++ )
		fprintf(file, " %.2i %.2i |", tel->Key.toState->Num, tel->Key.onChar );
	fprintf(file, "\n");
}

template < class State, class Key, class Transition >
		void Graph<State, Key, Transition>::
		Dump( FILE *file )
{
	/* Set the state numbers so we print something nice! */
	SetStateNumbers();

	/* Walk the list of states, printing. */
	State *state = Head;
	while (state) {
		fprintf( file, "%.2i\n", state->Num );
		DumpTransitions(file, &state->OutList);
		state = state->Next;
	}
}

#endif /* _AAPL_GRAPH_CC */
