/*
 *  Copyright 2001 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

%{

#define YY_NO_UNPUT

#include <iostream>
#include "rl.h"
#include "fsm.h"
#include "parsetree.h"
#include "rlparse.h"
#include "vector.h"

Vector<char> rlbuf;
int rlCurLine = 1;
int builtinBrace = 0;

%}

%x		OC_SGL_LIT
%x		OC_DBL_LIT
%x		OC_C_COM
%x		OC_C_COM_STAR
%x		OC_CXX_COM

%x		RL_INITIAL
%x		RL_SLIT
%x		RL_DLIT
%x		RL_ORLIT
%x		RL_REGEXP
%x		RL_REGEXP_OR
%x		RL_C_COM
%x		RL_C_COM_STAR
%x		RL_CXX_COM

%x		IL_INITIAL
%x		IL_SGL_LIT
%x		IL_DBL_LIT
%x		IL_C_COM
%x		IL_C_COM_STAR
%x		IL_CXX_COM

%%

	/* Numbers in outter code. */
<INITIAL>[0-9]+ {
	rllval.data = StrDup(rltext);
	return PP_Integer;
}
	/* Words in outter code. */
<INITIAL>[A-Za-z_][A-Za-z_0-9]* {
	rllval.data = StrDup(rltext);
	return PP_Word;
}

	/* Begin a c style comment. */
<INITIAL>"/*" {
	yymore();
	BEGIN(OC_C_COM);
}
	/* Data in a c style comment. */
<OC_C_COM>[^*\n] {
	yymore();
}
	/* Data in a c style comment. */
<OC_C_COM>\n {
	yymore();
	rlCurLine++;
}
	/* Move to a c comment star state. */
<OC_C_COM>"*" {
	yymore();
	BEGIN(OC_C_COM_STAR);
}
	/* Move back to a c comment state. */
<OC_C_COM_STAR>[^*/\n] {
	yymore();
	BEGIN(OC_C_COM);
}
	/* Move back to a c comment state. */
<OC_C_COM_STAR>\n {
	yymore();
	rlCurLine++; 
	BEGIN(OC_C_COM);
}
	/* Stay in c comment star state. */
<OC_C_COM_STAR>"*" {
	yymore();
}
	/* Terminate a c style comment. */
<OC_C_COM_STAR>"/" {
	BEGIN(INITIAL);
	rllval.data = StrDup(rltext);
	return PP_Comment;
}

	/* Begin a c++ style comment. */
<INITIAL>"//" {
	yymore();
	BEGIN(OC_CXX_COM);
}
	/* Data in a c++ style comment. */
<OC_CXX_COM>[^\n] {
	yymore();
}
	/* Terminalte a c++ style comment. */
<OC_CXX_COM>\n {
	rlCurLine++;
	BEGIN(INITIAL);
	rllval.data = StrDup(rltext);
	return PP_Comment;
}


	/* Start literals. */
<INITIAL>' {
	yymore();
	BEGIN(OC_SGL_LIT);
}
<INITIAL>\" {
	yymore();
	BEGIN(OC_DBL_LIT);
}
	/* Various escape sequences in literals. We don't need to get them
	 * all here. We just need to pick off the ones that could confuse us
	 * about the literal we are matchine */
<OC_SGL_LIT,OC_DBL_LIT>\\'		yymore();
<OC_SGL_LIT,OC_DBL_LIT>\\\"		yymore();
<OC_SGL_LIT,OC_DBL_LIT>\\\\		yymore();
	/* Characters in literals. */
<OC_DBL_LIT>[^\"\n]				yymore();
<OC_SGL_LIT>[^'\n]				yymore();
	/* Newlines in literals. */
<OC_SGL_LIT,OC_DBL_LIT>\n {
	yymore();
	rlCurLine++;
}
	/* Terminate a double literal */
<OC_DBL_LIT>\" {
	BEGIN(INITIAL);
	rllval.data = StrDup(rltext);
	return PP_Literal;
}
	/* Terminate a single literal. */
<OC_SGL_LIT>' {
	BEGIN(INITIAL);
	rllval.data = StrDup(rltext);
	return PP_Literal;
}

	/* Newline. */
<INITIAL>\n {
	rlCurLine++;
	rllval.data = StrDup(rltext);
	return PP_Whitespace;
}
<INITIAL>[\t ]+ {
	rllval.data = StrDup(rltext);
	return PP_Whitespace;
}

	/* Section Deliminator */
<INITIAL>"%%" {
	BEGIN(RL_INITIAL);
	return PP_Section;
}

	/* Any other characters. */
<INITIAL>. {
	rllval.data = StrDup(rltext);
	return PP_Character;
}

	/* End of input in a literal is an error. */
<OC_SGL_LIT,OC_DBL_LIT><<EOF>>		{
	fprintf(stderr, "%s:%i: unterminated literal\n",
			inputFile, rlCurLine );
	exit(1);
}
	
	/* End of input in a comment is an error. */
<OC_C_COM,OC_C_COM_STAR,OC_CXX_COM><<EOF>>	{
	fprintf(stderr, "%s:%i: unterminated comment\n",
			inputFile, rlCurLine );
	exit(1);
}


<RL_INITIAL>[0-9][0-9]* {	
	rllval.data = StrDup( rltext );
	return RL_Integer;
}
<RL_INITIAL>0x[0-9a-fA-F][0-9a-fA-F]* {	
	rllval.data = StrDup( rltext );
	return RL_Hex;
}
<RL_INITIAL>clear {
	rllval.data = StrDup( rltext );
	return RL_Clear;
}
<RL_INITIAL>init {
	rllval.data = StrDup( rltext );
	return RL_Init;
}
<RL_INITIAL>func {
	rllval.data = StrDup( rltext );
	return RL_Func;
}
<RL_INITIAL>prefunc {
	rllval.data = StrDup( rltext );
	return RL_PreFunc;
}
<RL_INITIAL>postfunc {
	rllval.data = StrDup( rltext );
	return RL_PostFunc;
}
<RL_INITIAL>struct {
	rllval.data = StrDup( rltext );
	return RL_Data;
}
<RL_INITIAL>[A-Za-z_][A-Za-z_0-9]* {
	rllval.data = StrDup( rltext );
	return RL_Word;
}

	/* Begin a c++ style comment. */
<RL_INITIAL>#					yymore(); BEGIN(RL_CXX_COM);
	/* Data in a c++ style comment. */
<RL_CXX_COM>[^\n]				yymore();
	/* Terminalte a c++ style comment. */
<RL_CXX_COM>\n					rlCurLine++; BEGIN(RL_INITIAL);

	/* Start literals. */
<RL_INITIAL>'					BEGIN(RL_SLIT);
<RL_INITIAL>\"					BEGIN(RL_DLIT);
<RL_INITIAL>"["                 BEGIN(RL_ORLIT);
<RL_SLIT,RL_DLIT,RL_ORLIT>\\0		rlbuf.Append('\0');
<RL_SLIT,RL_DLIT,RL_ORLIT>\\a		rlbuf.Append('\a');
<RL_SLIT,RL_DLIT,RL_ORLIT>\\b		rlbuf.Append('\b');
<RL_SLIT,RL_DLIT,RL_ORLIT>\\t		rlbuf.Append('\t');
<RL_SLIT,RL_DLIT,RL_ORLIT>\\n		rlbuf.Append('\n');
<RL_SLIT,RL_DLIT,RL_ORLIT>\\v		rlbuf.Append('\v');
<RL_SLIT,RL_DLIT,RL_ORLIT>\\f		rlbuf.Append('\f');
<RL_SLIT,RL_DLIT,RL_ORLIT>\\r		rlbuf.Append('\r');
<RL_SLIT,RL_DLIT,RL_ORLIT>\\\n		{ }
<RL_SLIT,RL_DLIT,RL_ORLIT>\\.		rlbuf.Append(rltext[1]);
	/* Characters in literals. */
<RL_SLIT>[^\\'\n]			rlbuf.Append(*rltext);
<RL_DLIT>[^\\"\n]			rlbuf.Append(*rltext);
<RL_ORLIT>[^\\\]\n]			rlbuf.Append(*rltext);
	/* Newlines in literals. */
<RL_SLIT,RL_DLIT,RL_ORLIT>\n {
	rlCurLine++;
	rlbuf.Append(*rltext);
}

	/* Terminate a single literal. */
<RL_SLIT>' {
	rlbuf.Append( 0 );
	rllval.data = StrDup( rlbuf.Table );
	rlbuf.Empty();
	BEGIN(RL_INITIAL);
	return RL_Literal;
}
	/* Terminate a double literal */
<RL_DLIT>\" {
	rlbuf.Append( 0 );
	rllval.data = StrDup( rlbuf.Table );
	rlbuf.Empty();
	BEGIN(RL_INITIAL);
	return RL_Literal;
}
	/* Terminate an or literal */
<RL_ORLIT>\] {
	rlbuf.Append( 0 );
	rllval.data = StrDup( rlbuf.Table );
	rlbuf.Empty();
	BEGIN(RL_INITIAL);
	return RL_OrLiteral;
}

<RL_INITIAL>\/		{
	BEGIN(RL_REGEXP);
	return RL_ReLiteralSlash;
}
<RL_REGEXP>"\[^"	{
	/*rlbuf.Append(RC_OpenNeg);*/
	BEGIN(RL_REGEXP_OR);
	return RL_ReLiteralOpenNeg;
}
<RL_REGEXP>"\["		{
	BEGIN(RL_REGEXP_OR);
	return RL_ReLiteralOpen;
}
<RL_REGEXP>\.		return RL_ReLiteralDot;
<RL_REGEXP>\*		return RL_ReLiteralStar;
<RL_REGEXP,RL_REGEXP_OR>\\0		rllval.chr = '\0'; return RL_ReLiteralChar;
<RL_REGEXP,RL_REGEXP_OR>\\a		rllval.chr = '\a'; return RL_ReLiteralChar;
<RL_REGEXP,RL_REGEXP_OR>\\b		rllval.chr = '\b'; return RL_ReLiteralChar;
<RL_REGEXP,RL_REGEXP_OR>\\t		rllval.chr = '\t'; return RL_ReLiteralChar;
<RL_REGEXP,RL_REGEXP_OR>\\n		rllval.chr = '\n'; return RL_ReLiteralChar;
<RL_REGEXP,RL_REGEXP_OR>\\v		rllval.chr = '\v'; return RL_ReLiteralChar;
<RL_REGEXP,RL_REGEXP_OR>\\f		rllval.chr = '\f'; return RL_ReLiteralChar;
<RL_REGEXP,RL_REGEXP_OR>\\r		rllval.chr = '\r'; return RL_ReLiteralChar;
<RL_REGEXP,RL_REGEXP_OR>\\\n		{ }
<RL_REGEXP>\\.		rllval.chr = rltext[1]; return RL_ReLiteralChar;
	/* Characters in literals. */
<RL_REGEXP>[^\\/\n]			rllval.chr = rltext[0]; return RL_ReLiteralChar;
	/* Newlines in literals. */
<RL_REGEXP>\n {
	rlCurLine++;
	rllval.chr = rltext[0];
	return RL_ReLiteralChar;
}

<RL_REGEXP_OR>.-[^\]]	{
	rllval.regExpSet.lower = rltext[0];
	rllval.regExpSet.upper = rltext[2];
	return RL_ReLiteralSet;
}
<RL_REGEXP_OR>\]	BEGIN(RL_REGEXP); return RL_ReLiteralClose;
<RL_REGEXP_OR>\\.	rllval.chr = rltext[1]; return RL_ReLiteralChar;

	/* Terminate a regular expression */
<RL_REGEXP>\/ {
	BEGIN(RL_INITIAL);
	return RL_ReLiteralSlash;
}
	/* Terminate an or literal */
<RL_REGEXP_OR>\/ {
	BEGIN(RL_INITIAL);
	return RL_ReLiteralSlash;
}

<RL_REGEXP_OR>.		rllval.chr = rltext[0]; return RL_ReLiteralChar;

	/* Anything else found in and/or/regexp literals. */
<RL_REGEXP,RL_REGEXP_OR,RL_SLIT,RL_DLIT,RL_ORLIT>.	{
	fprintf(stderr, "%s:%i: bad character"
			" in literal\n", inputFile, rlCurLine );
	exit(1);
}

	/* Whitespace. */
<RL_INITIAL>[\t ] { }
	/* Newline. */
<RL_INITIAL>\n {
	rlCurLine++;
}
	
	/* Builtin code move to Builtin initial. */
<RL_INITIAL>"{" {
	builtinBrace++;
	BEGIN(IL_INITIAL);
}

	/* Section Deliminator */
<RL_INITIAL>"%%" {
	BEGIN(INITIAL);
	return PP_Section;
}

	/* The double dot, we interpret this in the scanner becuase an
	 * LA-LR(1) parser cannot handle it, need LR(2). */
<RL_INITIAL>\.\. {
	return RL_DoubleDot;
}

	/* Any other characters. */
<RL_INITIAL>. {
	rllval.data = StrDup( rltext );
	return *rltext;
}


	/* End of input in a literal is an error. */
<RL_SLIT,RL_DLIT><<EOF>> {
	fprintf(stderr, "%s:%i: unterminated literal\n", 
			inputFile, rlCurLine );
	exit(1);
}
	
	/* End of input in a comment is an error. */
<RL_C_COM,RL_C_COM_STAR,RL_CXX_COM><<EOF>> {
	fprintf(stderr, "%s:%i: unterminated comment\n",
			inputFile, rlCurLine );
	exit(1);
}

	/* Begin a c style comment. */
<IL_INITIAL>"/*" {
	yymore();
	BEGIN(IL_C_COM);
}
	/* Data in a c style comment. */
<IL_C_COM>[^*\n] {
	yymore();
}
	/* Data in a c style comment. */
<IL_C_COM>\n {
	yymore(); 
	rlCurLine++;
}
	/* Move to a c comment star state. */
<IL_C_COM>"*" {
	yymore();
	BEGIN(IL_C_COM_STAR);
}
	/* Move back to a c comment state. */
<IL_C_COM_STAR>[^*/\n] {
	yymore();
	BEGIN(IL_C_COM);
}
	/* Move back to a c comment state. */
<IL_C_COM_STAR>\n {
	yymore(); 
	rlCurLine++; 
	BEGIN(IL_C_COM);
}
	/* Stay in c comment star state. */
<IL_C_COM_STAR>"*" {
	yymore();
}
	/* Terminate a c style comment. */
<IL_C_COM_STAR>"/" {
	yymore();
	BEGIN(IL_INITIAL);
}

	/* Begin a c++ style comment. */
<IL_INITIAL>"//" {
	yymore();
	BEGIN(IL_CXX_COM);
}
	/* Data in a c++ style comment. */
<IL_CXX_COM>[^\n] {
	yymore();
}
	/* Terminalte a c++ style comment. */
<IL_CXX_COM>\n {
	yymore();
	rlCurLine++;
	BEGIN(IL_INITIAL);
}


	/* Start literals. */
<IL_INITIAL>' {
	yymore();
	BEGIN(IL_SGL_LIT);
}
<IL_INITIAL>\" {
	yymore();
	BEGIN(IL_DBL_LIT);
}
	/* Various escape sequences in literals. We don't need to get them
	 * all here. We just need to pick off the ones that could confuse us
	 * about the literal we are matchine */
<IL_SGL_LIT,IL_DBL_LIT>\\'		yymore();
<IL_SGL_LIT,IL_DBL_LIT>\\\"		yymore();
<IL_SGL_LIT,IL_DBL_LIT>\\\\		yymore();
	/* Characters in literals. */
<IL_DBL_LIT>[^\"\n]				yymore();
<IL_SGL_LIT>[^'\n]				yymore();
	/* Newlines in literals. */
<IL_SGL_LIT,IL_DBL_LIT>\n {
	yymore();
	rlCurLine++;
}
	/* Terminate a double literal */
<IL_DBL_LIT>\" {
	yymore();
	BEGIN(IL_INITIAL);
}
	/* Terminate a single literal. */
<IL_SGL_LIT>' {
	yymore();
	BEGIN(IL_INITIAL);
}

	/* Open Brace, increment count of open braces. */
<IL_INITIAL>"{" {
	yymore();
	builtinBrace++;
}

	/* Close Brace, decrement count of open braces. */
<IL_INITIAL>"}" {
	builtinBrace--;
	if ( builtinBrace == 0 ) {
		BEGIN(RL_INITIAL);
		rllval.data = StrDup( rltext );
		// Remove the closing brace from the string. 
		rllval.data[ strlen(rllval.data) - 1 ] = 0;
		return RL_Builtin;
	}
	else {
		yymore();
	}
}

	/* Newline. */
<IL_INITIAL>\n {
	yymore();
	rlCurLine++;
}
	/* Any other characters. */
<IL_INITIAL>. {
	yymore();
}

	/* End of input in a literal is an error. */
<IL_SGL_LIT,IL_DBL_LIT><<EOF>>		{
	fprintf(stderr, "%s:%i: unterminated literal\n",
			inputFile, rlCurLine );
	exit(1);
}
	
	/* End of input in a comment is an error. */
<IL_C_COM,IL_C_COM_STAR,IL_CXX_COM><<EOF>>	{
	fprintf(stderr, "%s:%i: unterminated comment\n",
			inputFile, rlCurLine );
	exit(1);
}

%%

int yywrap()
{
	/* Once processessing of the input is done, no more. */
	return 1;
}

